#include "powerclass.h"


/**** NAND *****/
/**** Using SINGLESTACK and MULTSTACK principles *****/
float circuit::nandleakage(int fvalues[],int gid){
	int stim[0];
	int i,k=0;
	float leakage=0;
	int fancount;

	if(fanin[gid] > 4)                                        /***** doing this because benchmark circuits have simple logic gates with fanins greater than 4******/
	fancount=4;
	else
	fancount=fanin[gid];

	for(i=0; i < fancount; i++){
		if(fvalues[i] == 0){
			k++;
		}else if(fvalues[i] == 2){
			return 0;
		}
	}

    if(k == 0){
		stim[0]=fvalues[0];
		leakage = fanin[gid]*singlestackleak(stim,gid);
	}else{
		leakage= multistackleak(fvalues,gid,"n");
	}

	return leakage;
}

/***** AND : NAND + INV *****/
float circuit::andleakage(int fvalues[],int gid){
	int stim[0];

	float leakage=0;

	leakage=nandleakage(fvalues,gid);
	stim[0]=nandop(fvalues,fanin[gid]);

    if(leakage != 0)                                     /*** if leakage is zero means one input is X *****/
	leakage=leakage + notleakage(stim,0);

	return leakage;
}

/**** NOR *****/
float circuit::norleakage(int fvalues[],int gid){
	int stim[0];
	int i,k=0;
	float leakage=0;
	int fancount;

	if(fanin[gid] > 4)
		fancount=4;
	else
		fancount=fanin[gid];

	for(i=0; i < fancount; i++){
		if(fvalues[i] == 1){
			k++;
		}else if(fvalues[i] == 2){
			return 0;
		}
	}

    if(k == 0){
		stim[0]=fvalues[0];
		leakage = fanin[gid]*singlestackleak(stim,gid);
	}else{
		leakage= multistackleak(fvalues,gid,"p");
	}

	return leakage;
}

/**** OR : NOR + INV ****/
float circuit::orleakage(int fvalues[],int gid){
	int stim[0];
	float leakage=0;

	leakage=norleakage(fvalues,gid);
	stim[0]=norop(fvalues,fanin[gid]);

    if(leakage != 0)                                     /*** if leakage is zero means one input is X *****/
	leakage=leakage+notleakage(stim,0);

	return leakage;
}

/***** INV ******/
float circuit::notleakage(int fvalues[],int gid){
	if(fvalues[0] == 2)
	return 0;
	else
	return singlestackleak(fvalues,gid);
}

/**** BUF *****/
float circuit::buffleakage(int gid){
	int stim[0];
	float leakage=0;
	stim[0]=1;
	leakage=singlestackleak(stim,gid);
	stim[0]=0;
	leakage=leakage + singlestackleak(stim,gid);
	return leakage;
}

/**** XNOR ****/
float circuit::xnorleakage(int fvalues[],int gid){
    int stim[0];
    int i;
	float leakage=0;
	if(fvalues[0] == 0 && fvalues [1] ==0 ){
		leakage = multistackleak(fvalues,gid,"n");
		stim[0]=fvalues[0];
		leakage = leakage + 2*singlestackleak(stim,gid);
	}else if(fvalues[0] == 1 && fvalues[1] == 1){
		stim[0]=1;
		leakage= 2*singlestackleak(stim,gid);
		stim[0]=0;
		leakage=leakage+singlestackleak(stim,gid);
	}else if((fvalues[0] == 0 && fvalues[1] == 1) || (fvalues[0] == 1 && fvalues[1] == 0)){
		stim[0]=0;
		leakage = multistackleak(fvalues,gid,"n");
		stim[0]=1;
		leakage = leakage + multistackleak(fvalues,gid,"p") + singlestackleak(stim,gid);
	}else{
		return 0;
	}
	return leakage;
}

/**** XOR : XNOR + INV ****/
float circuit::xorleakage(int fvalues[],int gid){
    int stim[0];
	float leakage=0;

	leakage=xnorleakage(fvalues,gid);
	stim[0]=xnorop(fvalues,fanin[gid]);

    if(leakage != 0)                                     /*** if leakage is zero means one input is X *****/
	leakage=leakage+notleakage(stim,gid);

	return leakage;
}

/**** OAI *****/
float circuit::oaileakage(int faninvalues[],int gid){
    float leakage=0;
	int i;
    for(i=0;i<fanin[gid];i++){
		if(faninvalues[i] == 2)
		return 0;
	}

    leakage=readleakfile(gid,faninvalues,fanin[gid],"x");

    if(leakage == 0){
	   getcellsize(gid);
	   runcmplx(gid,faninvalues);
  	   leakage=readleakfile(gid,faninvalues,fanin[gid],"x");
   }

   return leakage;

}

/***** MUX ******/
float circuit::muxleakage(int faninvalues[],int gid){

	float leakage=0;
		int i;
	    for(i=0;i<fanin[gid];i++){
			if(faninvalues[i] == 2)
			return 0;
		}

	    leakage=readleakfile(gid,faninvalues,fanin[gid],"x");

	    if(leakage == 0){
		   getcellsize(gid);
		   runcmplx(gid,faninvalues);
	  	   leakage=readleakfile(gid,faninvalues,fanin[gid],"x");
	   }

	   return leakage;


}

/****** AOI ******/
float circuit::aoileakage(int faninvalues[],int gid){
     int i;
	 float leakage=0;

	 	for(i=0;i<fanin[gid];i++){
	 		if(faninvalues[i] == 2)
	 			return 0;
	     }



	 leakage=readleakfile(gid,faninvalues,fanin[gid],"x");

	    if(leakage == 0){
		   getcellsize(gid);
		   runcmplx(gid,faninvalues);
	  	   leakage=readleakfile(gid,faninvalues,fanin[gid],"x");
	   }

	 return leakage;

}

/****** AO : AOI + INV *****/
float circuit::aoleakage(int faninvalues[],int gid){

     int stim[0];
	 float leakage=0;
	 leakage=aoileakage(faninvalues,gid);


	 stim[0]=notop(aoop(faninvalues,gid));

	 if(leakage != 0)                                     /*** if leakage is zero means one input is X *****/
	 leakage=leakage+notleakage(stim,0);

	 return leakage;

}

/****** LATCH (same for latchr and latchs ignoring reset and set lines ******/
float circuit::latchleakage(int faninvalues[],int gid){
     int i;
	 float leakage=0;

	 	for(i=0;i<fanin[gid];i++){
	 		if(faninvalues[i] == 2)
	 			return 0;
	     }



	 leakage=readleakfile(gid,faninvalues,2,"x");                      /**** read in only D and clk inputs ****/

	    if(leakage == 0){
		   runlatch(gid,faninvalues);
	  	   leakage=readleakfile(gid,faninvalues,2,"x");
	   }

	 return leakage;

}


/**** DFF and DFFR (leakage estimated the same ;ignoring reset value at the moment) ****/

float circuit::dffleakage(int gid){
float leakunit=0;
int i;
int *faninvalues;

faninvalues=new int[fanin[gid]];


//cellpowerunit=0;

for(i=0;i<fanin[gid];i++){
	faninvalues[i]=gold[faninlist[gid][i]];
}

if(faninvalues[0] == 2)
return 0;

leakunit=readleakfile(gid,faninvalues,1,"0") + readleakfile(gid,faninvalues,1,"1");



if(leakunit == 0){
runff(gid,faninvalues);
leakunit=readleakfile(gid,faninvalues,1,"0") + readleakfile(gid,faninvalues,1,"1");
}

//cellpowerunit=readleakfile(gid,faninvalues,fanin[gid],"r") + readleakfile(gid,faninvalues,fanin[gid],"f");

return leakunit;

}

/**** DFF cell power (power due to internal switching)****/
float circuit::dffcellpower(int gid){

float pwrunit=0;
int i;
int *faninvalues;

faninvalues=new int[fanin[gid]];


for(i=0;i<fanin[gid];i++){
	faninvalues[i]=gold[faninlist[gid][i]];
}

if(faninvalues[0] == 2)
return 0;

pwrunit=readleakfile(gid,faninvalues,1,"r") + readleakfile(gid,faninvalues,1,"f");

return pwrunit;

}

/**** SINGLESTACK ****/
float circuit::singlestackleak(int fvalues[],int gateid){
float leak=0;

leak=readleakfile(gateid,fvalues,1,"x");

if(leak == 0){
  getcellsize(gateid);
  runscript(gateid,fvalues,1,"x");
  leak=readleakfile(gateid,fvalues,1,"x");
}

return leak;

}

/**** MULTISTACK ****/
float circuit::multistackleak(int fvalues[],int gateid,const char* type){
float leak=0;
int i;

leak=readleakfile(gateid,fvalues,fanin[gateid],type);

if(leak == 0){
  getcellsize(gateid);
  runscript(gateid,fvalues,fanin[gateid],type);
  leak=readleakfile(gateid,fvalues,fanin[gateid],type);
}

return leak;
}

float circuit::readleakfile(int gateid,int fvalues[],int fancount,const char* type){
FILE *infile2;
int i;
char fname[40],str1[50],str[2];
float leak=0;
sprintf(str, "%d",gtype[gateid]);
strcpy(fname,"models/pwr/");
strcat(fname,str);
strcat(fname,"_0");
sprintf(str, "%d",fanin[gateid]);
strcat(fname,str);
strcat(fname,".pwr");
infile2=fopen(fname,"r");

sprintf(str,"%d",fvalues[0]);
strcpy(str1,str);

if(fancount > 4 && (gtype[gateid] == NAND || gtype[gateid] == NAND_2X || gtype[gateid] == NOR || gtype[gateid] == NOR_2X || gtype[gateid] == NOR_ii || gtype[gateid] == AND || gtype[gateid] == OR))
fancount=4;

for(i=1;i<fancount;i++){
	sprintf(str,"%d",fvalues[i]);
	strcat(str1,str);

}

if(strcmp(type,"x")){
	strcat(str1,":");
	strcat(str1,type);
}


strcat(str1,":leakage=");

/*if(gtype[gateid] == AND && fanin[gateid] == 5)
printf("gateid %d str1 %s \n",gateid,str1);*/


if(infile2 == NULL){
return 0;
}else{
	   while(!feof(infile2)){
		   fscanf(infile2,"%s",&fname);
			if(!strcmp(fname,str1)){
				fscanf(infile2,"%f",&leak);
				fclose(infile2);
				return leak;
			}
		}
	}
fclose(infile2);
return 0;
}



