#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <math.h>

enum
{
   JUNK ,        /* 0 */
   INPUT,        /* 1 */
   OUTPUT,       /* 2 */
   AND,          /* 3 */
   NAND,         /* 4 */
   OR,           /* 5 */
   NOR,          /* 6 */
   NOR_ii,       /* 7 */
   INV,          /* 8 */
   INV_2X,       /* 9 */
   BUFF,         /* 10 */
   XOR,          /* 11 */
   XOR_2X,       /* 12 */
   XNOR,         /* 13 */
   XNOR_2X,      /* 14 */
   AO21,         /* 15 */
   AO22,         /* 16 */
   AO221,        /* 17 */
   AO32,         /* 18 */
   AOI21,        /* 19 */
   AOI22,        /* 20 */
   AOI221,       /* 21 */
   AOI222,       /* 22 */
   AOI32,        /* 23 */
   AOI321,       /* 24 */
   AOI322,       /* 25 */
   AOI33,        /* 26 */
   AOI43,        /* 27 */
   OAI21,        /* 28 */
   OAI22,        /* 29 */
   OAI221,       /* 30 */
   OAI222,       /* 31 */
   OAI32,        /* 32 */
   OAI321,       /* 33 */
   OAI322,       /* 34 */
   OAI33 ,       /* 35 */
   OAI43,        /* 36 */
   OAI332,       /* 37 */
   OAI44,        /* 38 */
   MUX,			 /* 39 */
   MUX_NI,		 /* 40 */
   DFF,          /* 41 */
   DFFR,		 /* 42 */
   QBAR,         /* 43 */
   LATCH,        /* 44 */
   NAND_2X,      /* 45 */
   NOR_2X,       /* 46 */
   BUFF_2X,      /* 47 */
   CLK,			 /* 48 */
   LATCHR,       /* 49 */
   LATCHS,       /* 50 */
   LATCHS_NI,    /* 51 */
   INV_4X,       /* 52 */
   BUFF_4X,      /* 53 */
   INV_8X,       /* 54 */
   BUFF_8X,      /* 55 */
   INV_12X,      /* 56 */
   BUFF_12X,     /* 57 */
   INV_16X,      /* 58 */
};

#define INFI	32767

class circuit{

    int fid;
	int ipcount,gatecount,linecount;
	//int gateid,gid,gtid,gsid;
	int **eventlist,**scheduled,**gscheduled,**badevents;
	int **gateevent;
	int *listnum,*processed,*glistnum,*blistnum;
	/*int *level;*/
	int **fanoutlist,**faninlist;
	int *gtype;
	int *gold,**gnew,*nextime;
	int *gevent;                              /*events occuring on that particular gate*/
	int *gzero;
	float *trf;                              /*** denotes the rise or fall time of the  signal at the gate output node***/
	float *leakage,*cellpowerunit;
	int *fanin,*fanout;
	int time_wheel_size;
	int maxdelay;
	int chdelay,choice;
	float temp;
	int vecperiod,clkperiod;
	float energy;
	char techfile[40];
	float lambda;
	int Mn,Mp;    /*no of parallel transistors nmos/pmos*/
	int zerodelay;


	float Cox;// Cox = 6 F/m^2
	float CGDOp;// CGDOp+CGSOp
	float CGDOn;// CGDOn+CGSOn
	float CGBO; /*gate bulk capacitance*/
	float Wp;
	float Wn;
	float L;
	float Vdd;
	float TOX;
	float tr;
	float Ps,Psini,T;
	float ffcellpower;

	float isSeq ;
	int clktick;
	int clkvalue;                                          /* for latches */
	int clkid;                                             /* gate id of main clock*/
	int clkevent;                                          /* to depict edge triggering */
	int time;
	//float cellpowerunit;
	float e0,e1,e01,e11;                            /**** technology parameters for short circuit evaluation *****/
	float e0p,e1p,e2p,e0n,e1n,e2n;                            /**** technology parameters for short circuit evaluation *****/
	float Vtn,Vtp,Bn,Bp;

public:

circuit(char *,char *);  // constructor
~circuit();    //destructor
void simulation(char *cktname);
void getcellsize(int gid);
//void runscript(int g1,int stim);
void runscript(int g1,int fvalues[],int fanin,const char* type);
void runcmplx(int g1,int fvalues[]);
void runff(int g1,int fvalues[]);
void runlatch(int g1,int fvalues[]);
void randomgendelayfile(char *cktname);
int readdelayfile(char *cktname,int gateid);
int maxdelayfile(char *cktname);

float shortckt(int gateid,float cload, int time);
float Ipavg(float cload,float t0,char chk);
int glitchfilter(char *cktname,int gateid,int time);
int evaluate(int gid);
float leakagepow(int gid);
float getcload(int gateid);
int andop(int fvalues[],int count);
int orop(int fvalues[],int count);
int xorop(int fvalues[],int count);
int xnorop(int fvalues[],int count);
int nandop(int fvalues[],int count);
int oaiop(int fvalues[],int gateid);
int dffop(int fvalues[],int gateid);
int latchop(int fvalues[],int gateid);
int aoop(int fvalues[],int gateid);
int muxop(int fvalues[],int gateid);
int aoiop(int fvalues[],int gateid);
int notop(int fvalues);
int norop(int fvalues[],int count);
int norii_op(int fvalues[],int count);
int delay(char *cktname,int gateid);
int chkprocess(int *processed,int time);
int maxcal(int *fanout,int count);
float nandleakage(int fvalues[],int gid);
float oaileakage(int fvalues[],int gid);
float aoileakage(int fvalues[],int gid);
float aoleakage(int fvalues[],int gid);
float dffleakage(int gid);
float latchleakage(int faninvalues[],int gid);
float dffcellpower(int gid);
float muxleakage(int fvalues[],int gid);
float dffrleakage(int fvalues[],int gid);
float andleakage(int fvalues[],int gid);
float orleakage(int fvalues[],int gid);
float notleakage(int fvalues[],int gid);
float norleakage(int fvalues[],int gid);
float xorleakage(int fvalues[],int gid);
float xnorleakage(int fvalues[],int gid);
float buffleakage(int gid);
float readcapfile(int gateid,int gid,char *search);
//float getUeffp();
//float getUeffn();
//float getRon(char c);
float getcjunc(int gateid);

float singlestackleak(int fvalues[],int gateid);
float multistackleak(int fvalues[],int gateid,const char* type);
float readleakfile(int gateid,int fvalues[],int fanin,const char* type);
float pmosleakage();
float nmosleakage();
float readcaplist(char *cktname,int gateid);
void runstandmodel();
void runstand(int input,int flag1,int flag2);
void readtechlib();
void readbeta(int gateid,int input);
float notsckt(int gateid,int input ,int t,float cload,float trr);
float nandsckt(int gateid,int input ,int t,float cload,float trr);
float andsckt(int gateid,int input ,int t,float cload,float trr);
float norsckt(int gateid,int input ,int t,float cload,float trr);
float orsckt(int gateid,int input ,int t,float cload,float trr);
float xorsckt(int gateid,int input ,int t,float cload,float trr);
float xnorsckt(int gateid,int input ,int t,float cload,float trr);
float ao22sckt(int gateid,int input ,int t,float cload,float trr);
float bufsckt(int gateid,int input ,int t,float cload,float trr);
float muxsckt(int gateid,int input ,int t,float cload,float trr);
float Ipavg(int gateid,int time,float cload,float t0);
};
